# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the HasLinkOrPartOfEquationSet rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite
import scade.model.traceability.local_trace_model as local_trace
import scade.model.traceability.traceability as trace

from ansys.scade.design_rules.utils.modelling import get_model
from ansys.scade.design_rules.utils.rule import SCK, Rule


class HasLinkOrPartOfEquationSet(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0026',
        category='Traceability',
        severity=Rule.REQUIRED,
        kinds=None,
        description=(
            'This rule checks if the elements have traceability links '
            'via the ALM Gateway or is part of an Equation Set.'
        ),
        label='Has traceability link',
    ):
        if not kinds:
            kinds = [SCK.CONSTANT, SCK.OPERATOR, SCK.DIAGRAM, SCK.EQ_SET]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=None,
            kinds=kinds,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Load the traceability data."""
        # traceability dictionaries for each model/library
        self.almgr_dict = {}
        self.almgt_dict = {}
        for main_lib in model.all_libraries:
            path_name = main_lib.descriptor.model_file_name

            almgr_parser = trace.AlmgrParser()
            try:
                almgr_parser.parse(path_name)
                self.almgr_dict[main_lib] = almgr_parser
            except BaseException as e:
                self.set_message(f'almgr file could not be parsed ({e})')
                return Rule.ERROR

            almgt_parser = trace.AlmgtParser()
            try:
                almgt_parser.parse(path_name)
                self.almgt_dict[main_lib] = almgt_parser
            except BaseException as e:
                self.set_message(f'almgt file could not be parsed ({e})')
                return Rule.ERROR

        return Rule.OK

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # protect against kinds that are not applicable for this rule
        if not isinstance(object_, suite.Traceable):
            return Rule.NA

        if self._element_has_link(object_):
            has_link = True
        else:
            if isinstance(object_, suite.EquationSet):
                has_link = False
            else:
                has_link = self._is_part_of_equationset(object_)

        if not has_link:
            self.set_message('No linked requirement')
            return Rule.FAILED

        return Rule.OK

    def _element_has_link(self, element):
        model = get_model(element)
        # same check for libraries and main model
        almgr_parser = self.almgr_dict[model]
        almgt_parser = self.almgt_dict[model]

        number_of_links = 0
        # find links in almgr file
        for alm_document in almgr_parser.project.alm_documents:
            for requirement in alm_document.requirements:
                for incoming_link in requirement.incoming_links:
                    source = incoming_link.source
                    linked_object = model.get_object_from_oid(source.identifier)
                    if linked_object == element:
                        number_of_links += 1

        # update links with almgt information
        if almgt_parser.traceability:
            for link in almgt_parser.traceability.links:
                linked_object = model.get_object_from_oid(link.oid)
                if linked_object == element:
                    if link.action == local_trace.TraceType.ADD_LINK:
                        number_of_links += 1
                    elif link.action == local_trace.TraceType.UPDATE_LINK:
                        # TODO(Jean): needs to be checked
                        # https://github.com/ansys/scade-design-rules/issues/29
                        number_of_links += 1
                    elif link.action == local_trace.TraceType.REMOVE_LINK:
                        number_of_links -= 1

        return number_of_links > 0

    def _is_part_of_equationset(self, object_: suite.Object) -> bool:
        result = False
        if isinstance(object_, suite.Operator):
            for expr_call in object_.expr_calls:
                result |= self._has_equation_equation_set_with_req(expr_call.equation)
        elif (
            isinstance(object_, suite.LocalVariable) and (object_.is_input() or object_.is_hidden())
        ) or isinstance(object_, suite.Constant):
            for expr_id in object_.expr_ids:
                result |= self._has_equation_equation_set_with_req(expr_id.equation)
        elif isinstance(object_, suite.LocalVariable) and object_.is_output():
            for definition in object_.definitions:
                result |= self._has_equation_equation_set_with_req(definition)
        elif isinstance(object_, suite.LocalVariable):
            pass
        else:
            pass
        return result

    def _has_equation_equation_set_with_req(self, equation) -> bool:
        if equation is not None:
            for equation_set in equation.equation_sets:
                if self._element_has_link(equation_set):
                    return True
        return False


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    HasLinkOrPartOfEquationSet()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the AnnNotesForBasicInterfaceTypes rule."""

import scade.model.suite as suite

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from typing import List

from ansys.scade.apitools.query import get_cell_type, get_leaf_type, is_array
from ansys.scade.design_rules.utils.annotations import (
    AnnotationRule,
    ArgumentParser,
    get_first_note_by_type,
    is_ann_note_value_defined_and_get_value,
)
from ansys.scade.design_rules.utils.modelling import is_numeric, is_visible
from ansys.scade.design_rules.utils.rule import SCK, Rule


class AnnNotesForBasicInterfaceTypes(AnnotationRule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0099',
        category='annotations',
        severity=Rule.REQUIRED,
        parameter='-t SDD_TopLevel -v Public',
        description=(
            'All basic types used inside an operator interface shall have an annotation '
            'with the SI-Units, resolution, etc.\n'
            'NamedTypes are checked recursively.\n\n'
            'parameters:\n'
            "* '-t': Name of the annotation note type\n"
            "* '--public ': Public interfaces only (e.g.: '-t SDD_TopLevel --public')"
        ),
        label='AnnotationNotes for basic types in operator interface',
        # ease customization
        std_fields=None,
        numeric_fields=None,
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=None,
            kinds=[SCK.INPUT, SCK.OUTPUT, SCK.HIDDEN],
        )
        self.std_fields = std_fields if std_fields else ['Description', 'Constraints']
        self.numeric_fields = (
            numeric_fields if numeric_fields else ['Min_Value', 'Max_Value', 'Unit_SI']
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        # minimal level of backward compatibility
        parameter = parameter.replace(',visibility=Public', ' --public')
        parameter = parameter.replace(',visibility=Private', '')

        result = super().on_start(model, parameter)
        if result == Rule.OK:
            assert self.options is not None  # nosec B101  # addresses linter
            self.public = self.options.public
            # cache for checked named types of fields
            self.checked_typed = set()
        return result

    def add_arguments(self, parser: ArgumentParser):
        """Declare arguments in addition to the note type."""
        parser.add_argument('--public', help='Public interfaces only', action='store_true')

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # take care of sensors, should they be added later in the checked elements (kinds)
        declaration = object_.owner if isinstance(object_, suite.LocalVariable) else object_
        if not self.public or is_visible(declaration):
            self._check_object(object_)

        return Rule.NA

    def _check_object(self, typed: suite.TypedObject):
        # typed: context of the check, for annotations
        # must not be a table or a predefined type, that are typed objects without annotations
        # assert not isinstance(typed, suite.Table)
        # assert not isinstance(typed, suite.NamedType) or not typed.is_predefined()

        # use a cache to avoid duplicated analysis
        if typed in self.checked_typed:
            return
        self.checked_typed.add(typed)

        type_ = typed.type
        if is_array(type_):
            leaf = get_leaf_type(type_)
            # get the closest owner that is not a table: must be the leaf's owner
            typed = leaf.owner
            type_ = get_cell_type(leaf)

        if isinstance(type_, suite.NamedType) and type_.is_generic():
            # out of the rule's scope
            return
        elif isinstance(type_, suite.NamedType) and not type_.is_predefined():
            # bypass alias
            self._check_object(type_)
        elif isinstance(type_, suite.Structure):
            for element in type_.elements:
                # recurse
                self._check_object(element)
        else:
            # predefined, sized, imported, or enumeration types
            violation_text_missing = self._check_annotation(typed)
            if violation_text_missing:
                # TODO(Jean): object may not have a name
                # https://github.com/ansys/scade-design-rules/issues/29
                message = (
                    f'Annotation missing for {typed.name}: {", ".join(violation_text_missing)}'
                )
                self.add_rule_status(typed, Rule.FAILED, message, '')

    def _check_annotation(self, object_: suite.Annotable) -> List[str]:
        """Check the annotation for the given object."""
        violation_text_missing = []

        ann_note = get_first_note_by_type(object_, self.note_type)
        for field in self.std_fields:
            defined, _ = is_ann_note_value_defined_and_get_value(ann_note, field)
            if not defined:
                violation_text_missing.append(field)
        if is_numeric(object_.type) or (
            isinstance(object_, suite.NamedType) and is_numeric(object_)
        ):
            for field in self.numeric_fields:
                defined, _ = is_ann_note_value_defined_and_get_value(ann_note, field)
                if not defined:
                    violation_text_missing.append(field)

        return violation_text_missing


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    AnnNotesForBasicInterfaceTypes()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the UserDefinedTypesLib rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.apitools.query import get_type_name, is_array, is_structure
from ansys.scade.design_rules.utils.modelling import get_model
from ansys.scade.design_rules.utils.rule import SCK, Rule


class UserDefinedTypesLib(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0090',
        category='Project Structure',
        severity=Rule.REQUIRED,
        parameter='Domain',
        description=(
            'This rule checks if user-defined complex types used in top-level operators as '
            "interface are located in library projects prefixed with 'Lib' and the Domain name."
        ),
        label=(
            'Check that user-defined complex types used in operators '
            'at top-level are located in libraries'
        ),
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=None,
            kinds=[SCK.INPUT, SCK.HIDDEN, SCK.OUTPUT],
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        if model.name.startswith('Lib'):
            self.set_message(
                'Provided model is located in a library project. '
                'This rule cannot check library projects!'
            )
            return Rule.ERROR

        return Rule.OK

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        message = ''
        return_status = Rule.OK

        operator = object_.owner
        # Check if operator is at top-level: <model>/<package>::<operator>
        if not isinstance(operator.owner.owner, suite.Model):
            return Rule.NA

        model_name = operator.owner.owner.name
        type_ = object_.type
        if is_array(type_) or is_structure(type_):
            model = get_model(object_.type)
            assert model is not None  # nosec B101  # addresses linter
            if model.name == model_name:
                message = (
                    get_type_name(type_) + ' is user-defined complex type at top-level. '
                    'Needs to be moved to a Library.'
                )
                return_status = Rule.FAILED
            elif model.name != 'Lib' + parameter:
                message = (
                    get_type_name(type_)
                    + ' is user-defined complex type which is not located in Lib'
                    + parameter
                    + ' project.'
                )
                return_status = Rule.FAILED

        self.set_message(message)
        return return_status


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    UserDefinedTypesLib()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utilities for developing rules."""

from enum import Enum
from typing import Union

import scade.model.suite as suite

import ansys.scade.apitools.expr as expr
from ansys.scade.apitools.expr import Eck
from ansys.scade.apitools.prop import get_pragma_tool_text


def get_model(object_: suite.Object) -> suite.Model:
    """
    Get the model in which the object resides.

    This function cannot be called for objects that are not
    defined in a model, such as predefined types, for example.

    Parameters
    ----------
    object_ : suite.Object
        Input object.

    Returns
    -------
    suite.Model
    """
    assert object_.defined_in is not None  # nosec B101  # addresses linter
    return object_.defined_in.model


def get_top_level_package(object_: suite.Object) -> suite.Package:
    """
    Get the top level package of an object.

    Parameters
    ----------
    object_ : suite.Object
        Input object.

    Returns
    -------
    suite.Package
        Top level package or the model if no package is available.
    """
    container = object_.owner
    while container and not isinstance(container, suite.Package):
        object_ = container
        container = container.owner
    if isinstance(object_, suite.Package):
        return object_
    else:
        return container


def get_owner_package(object_: suite.Object) -> suite.Package:
    """
    Get the package of an object.

    Parameters
    ----------
    object_ : suite.Object
        Input object.

    Returns
    -------
    suite.Package
        Package or the model if no package is available.
    """
    container = object_.owner
    while container and not isinstance(container, suite.Package):
        container = container.owner
    return container


def get_root_container_of_expression(expression: suite.Expression) -> suite.Object:
    """
    Get the root container of an expression.

    Parameters
    ----------
    expression : suite.Expression
        Input expression.

    Returns
    -------
    suite.Object
        First container of the expression that is not an expression.
    """
    container = expression.owner
    while isinstance(container, suite.Expression):
        container = container.owner
    return container


def get_root_container_of_type(type_: suite.Type) -> suite.Object:
    """
    Get the root container of a type.

    Parameters
    ----------
    type_ : suite.Type
        Input type.

    Returns
    -------
    suite.Object
    """
    container = type_.owner
    while isinstance(container, suite.Type):
        container = container.owner
    return container


def is_expression_constant(expression: suite.Expression) -> bool:
    """
    Return whether an expression is a constant.

    Parameters
    ----------
    expression : suite.Expression
        Input expression.

    Returns
    -------
    bool
    """
    if isinstance(expression, suite.ConstValue):
        return True
    elif isinstance(expression, suite.ExprId):
        reference = expression.reference
        return isinstance(reference, suite.Constant)
    else:
        return False


class IR(Enum):
    """Identifies the different roles a variable may have in an iterated call."""

    NONE, INDEX, ACC_IN, ACC_OUT, CONTINUE = range(5)


def get_iter_role(variable: suite.LocalVariable, call: suite.ExprCall) -> IR:
    """
    Return the role that an operator variable may have in an iterated call.

    Parameters
    ----------
    variable : suite.LocalVariable
        Input, hidden input or output of the called operator.
    call : suite.ExprCall
        Call expression to the operator.

    Returns
    -------
    IR
        Role of the variable for the call.
    """
    assert variable.operator is not None  # nosec B101  # addresses linter
    operator = call.operator
    # assert operator == variable.operator
    op = expr.accessor(call)
    if not isinstance(op, expr.IteratorOp):
        return IR.NONE

    # parameters depending on the various iterators
    ix_index = -1  # index of the index input
    ix_continue = -1  # index of the condition output

    # number of accumulators
    assert isinstance(op, expr.IteratorOp)  # nosec B101  # addresses linter
    count = op.accumulator_count
    if count:
        # the number of accumulator for mapfold* constructs must be a literal
        assert isinstance(count, expr.ConstValue)  # nosec B101  # addresses linter
        n = int(count.value)
    else:
        # unless specified hereafter for fold* constructs
        n = 0

    if op.code == Eck.MAPFOLD:
        pass
    elif op.code == Eck.MAPFOLDI:
        ix_index = 0
    elif op.code == Eck.MAPFOLDW:
        ix_continue = 0
    elif op.code == Eck.MAPFOLDWI:
        ix_index = 0
        ix_continue = 0
    elif op.code == Eck.FOLD:
        n = 1
    elif op.code == Eck.FOLDI:
        n = 1
        ix_index = 0
    elif op.code == Eck.FOLDW:
        n = 1
        ix_continue = 0
    elif op.code == Eck.FOLDWI:
        n = 1
        ix_index = 0
        ix_continue = 0
    elif op.code == Eck.MAP:
        pass
    elif op.code == Eck.MAPI:
        ix_index = 0
    elif op.code == Eck.MAPW:
        ix_continue = 0
    elif op.code == Eck.MAPWI:
        ix_index = 0
        ix_continue = 0

    # index of the first input accumulator
    ix_in_acc = ix_index + 1
    # index of the first output accumulator
    ix_out_acc = ix_continue + 1

    if variable.is_output():
        if ix_continue != -1 and variable == operator.outputs[ix_continue]:
            return IR.CONTINUE
        elif n > 0 and variable in operator.outputs[ix_out_acc : ix_out_acc + n]:
            return IR.ACC_OUT
    else:
        # assert variable.is_input() or variable.is_hidden()
        if ix_index != -1 and variable == (operator.inputs + operator.hiddens)[0]:
            return IR.INDEX
        elif n > 0 and variable in (operator.inputs + operator.hiddens)[ix_in_acc : ix_in_acc + n]:
            return IR.ACC_IN

    # default
    return IR.NONE


def get_full_path_ex(object_: suite.Object) -> str:
    """
    Provide a Scade path for objects that don't have one.

    For objects that don't have a path, the function returns their name
    appended to their owner's path. Otherwise, returns their own path.

    Parameters
    ----------
    object_ : suite.Object
        Input object.

    Returns
    -------
    str
        Scade path of the object.
    """
    # a few objects do not have a path
    if isinstance(object_, suite.CompositeElement):
        return f'{object_.owner.get_full_path()}{object_.name}'
    # default
    return object_.get_full_path()


def get_path(scope: suite.Object, element: suite.Object) -> str:
    """
    Return the relative Scade path of a model element with respect to a scope.

    This function provides a workaround for ``Object.get_path()`` which returns
    a wrong path in some circumstances, for example with state machines.

    Parameters
    ----------
    scope : suite.Object
        Context for computing the relative path.
    element : suite.Object
        Input element.

    Returns
    -------
    str
        Relative Scade path of the element or .its absolute path if it not relative to ``scope``.
    """
    scope_path = scope.get_full_path()
    element_path = element.get_full_path()
    return element_path.lstrip(scope_path)


def get_enum_value(enum: suite.Constant) -> int:
    """
    Get the value of the ``C:enum_val`` kcg pragma of an enumeration value.

    Parameters
    ----------
    enum : suite.Constant
        Input enumeration value.

    Returns
    -------
    int
        Value of the kcg pragma ``C:enum_val`` if present, otherwise -1.
    """
    value = get_pragma_tool_text(enum, 'kcg', 'C:enum_val')
    return int(value) if value else -1


def is_numeric(type_: suite.Type) -> bool:
    """
    Return whether a type is numeric.

    Parameters
    ----------
    type_ : Type
        Input type.

    Returns
    -------
    bool
    """
    if isinstance(type_, suite.NamedType):
        if type_.is_predefined():
            return type_.name != 'bool' and type_.name != 'char'
        elif type_.is_generic() or type_.is_imported():
            # all the existing constraints are numeric
            return type_.constraint is not None
        else:
            # unknown
            return False
    else:
        return isinstance(type_, suite.SizedType)


def string_to_integer_float(string: str) -> Union[int, float]:
    """
    Return the numeric value of a Scade literal, by removing its suffix.

    The suffix is _[u]i[8|16|32|64] for integers of _f[32|64]for floats.

    Parameters
    ----------
    string : str
        Input Scade literal.

    Returns
    -------
    int | float
        Value as integer or float, depending on the type of the literal.
    """
    value = string.split('_')[0]
    if '.' in value:
        return float(value)
    else:
        return int(value)


def get_value_from_const(expression: suite.Expression) -> Union[int, float]:
    """
    Calculate the value of a Constant or ConstValue.

    Parameters
    ----------
    expression : suite.Expression
        Input expression.

    Returns
    -------
    int | float
        Value as integer or float, depending on the type of the literal.
    """
    if isinstance(expression, suite.ConstValue):
        val = expression.get_value()
        return string_to_integer_float(val)
    elif isinstance(expression, suite.Constant):
        return get_value_from_const(expression.value)
    else:
        expr_val = expression.evaluate_expression()
        return string_to_integer_float(expr_val)


def is_visible(element: suite.StorageElement) -> bool:
    """
    Return whether a package's declaration is visible.

    Parameters
    ----------
    declaration : StorageElement
        Input declaration.

    Returns
    -------
    bool
        Whether the element is visible, as well as its containing packages.
    """
    return isinstance(element.owner, suite.Model) or (
        element.visibility == 'Public' and is_visible(element.owner)
    )



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ConstantIfThenElse rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

import ansys.scade.apitools.expr.access as access
from ansys.scade.apitools.query import get_leaf_type, is_predefined
from ansys.scade.design_rules.utils.rule import Rule


class ConstantIfThenElse(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0011',
        category='Modelling',
        severity=Rule.ADVISORY,
        description=(
            "This rule checks if predefined Operators 'if..then..else' have constant "
            'boolean inputs on the left inputs.'
        ),
        label='This operator should not be used with constant inputs of boolean type',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        call = access.accessor(object_)
        if isinstance(call, access.IfThenElseOp):
            for expr in [call.if_] + call.then + call.else_:
                if self._find_const_bool_expr(expr):
                    violated = True
                    break
            else:
                violated = False
            if violated:
                container = self.get_closest_annotatable(object_)
                error_msg = 'If..Then..Else with at least one constant boolean input.'
                identifier = object_.to_string()
                self.add_rule_status(container, Rule.FAILED, error_msg, identifier)

        return Rule.NA

    def _find_const_bool_expr(self, param: access.Expression) -> bool:
        if isinstance(param, access.ConstValue):
            return param.value in ['true', 'false']
        elif isinstance(param, access.IdExpression):
            const_check = self._find_const_bool_flow(param.path)
            return const_check
        else:
            return False

    def _find_const_bool_flow(self, var: suite.ConstVar) -> bool:
        if isinstance(var, suite.Constant):
            type_ = get_leaf_type(var.type)
            return is_predefined(type_) and type_.name == 'bool'
        elif not var.definitions:
            return False
        else:
            # return true is all the producers are boolean constants:
            for eq in var.definitions:
                if not self._find_const_bool_expr(access.accessor(eq.right)):
                    return False
            return True


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ConstantIfThenElse()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LocalNameUniqueness rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite
from scade.model.suite.visitors import Visit

from ansys.scade.design_rules.utils.rule import Rule


class _GatherLocalVariables(Visit):
    """Dictionary of local variables, indexed by name."""

    def __init__(self, operator: suite.Operator):
        self.variables = {}
        self.visit(operator)

    def visit_local_variable(self, local_variable: suite.LocalVariable, *args):
        if not local_variable.is_internal():
            self.variables.setdefault(local_variable.name, []).append(local_variable)


class LocalNameUniqueness(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0038',
        label='Local name uniqueness',
        description='The names of local variables and signals shall be unique within an operator.',
        category='Readability',
        severity=Rule.ADVISORY,
        **kwargs,
    ):
        # The rule applies to local variables but the computations are done at the operator level
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[suite.Operator, suite.LocalVariable],
            has_parameter=False,
            **kwargs,
        )
        self.cache_variables = {}

    def before_checking_subtree(self, object_: suite.Object, parameter: str = '') -> int:
        """Cache all the local variables when object_ is an operator."""
        if isinstance(object_, suite.Operator):
            self.cache_variables = _GatherLocalVariables(object_).variables
        return Rule.OK

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # default
        status = Rule.OK
        if isinstance(object_, suite.LocalVariable) and not object_.is_internal():
            variables = self.cache_variables[object_.name]
            if len(variables) > 1:
                message = f'{object_.name}: Not unique name'
                self.set_message(message)
                status = Rule.FAILED
        return status


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LocalNameUniqueness()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utilities for developing rules related to annotations."""

from typing import List, Optional, Tuple

import scade
import scade.model.suite.annotation as ann
import scade.model.suite.suite as suite

from ansys.scade.design_rules.utils.command import ArgumentParser, ParameterParser
from ansys.scade.design_rules.utils.rule import Rule


def get_note_type(model: suite.Model, name: str) -> Optional[ann.AnnNoteType]:
    """
    Get an annotation type by name, or ``None`` if not found.

    Parameters
    ----------
    model : suite.Model
        Input model.
    name : str
        Name of the note type to search.

    Returns
    -------
    ann.AnnNoteType
        Found note type, ``None`` otherwise.
    """
    return next((_ for _ in model.ann_note_types if _.name == name), None)


def get_notes_by_type(object_: suite.Annotable, type_: ann.AnnNoteType) -> List[ann.AnnNote]:
    """
    Get the annotation notes of an object by type.

    Parameters
    ----------
    object_ : suite.Annotable
        Input object.
    type_ : ann.AnnNoteType
        Type of the notes to search.

    Returns
    -------
    List[ann.AnnNote]
        Found notes.
    """
    return [_ for _ in object_.ann_notes if _.ann_note_type == type_]


def get_first_note_by_type(
    object_: suite.Annotable,
    type_: ann.AnnNoteType,
) -> Optional[ann.AnnNote]:
    """
    Get the first annotation note of an object by type.

    Parameters
    ----------
    object_ : suite.Annotable
        Input object.
    type_ : ann.AnnNoteType
        Type of the notes to search.

    Returns
    -------
    Optional[ann.AnnNote]
        Found note, otherwise ``None``.
    """
    notes = get_notes_by_type(object_, type_)
    return notes[0] if notes else None


def get_ann_note_by_name(object_: suite.Annotable, name: str) -> Optional[ann.AnnNote]:
    """
    Get the annotation note of an object by name, or ``None`` if not found.

    DEPRECATED: note names are not reliable, although it is unlikely

    Parameters
    ----------
    object_ : suite.Annotable
        Input object.
    name : str
        Name of the note to search.

    Returns
    -------
    Optional[ann.AnnNote]
        Found note, otherwise ``None``.
    """
    return next((_ for _ in object_.ann_notes if _.name == name), None)


def is_ann_note_value_defined_and_get_value(
    note: Optional[ann.AnnNote],
    element: str,
) -> Tuple[bool, str]:
    """
    Get the attribute value of a note and whether it is not empty.

    The attribute's value '-' is considered as empty.

    Parameters
    ----------
    note : Optional[ann.AnnNote]
        Input object.
    element : str
        Name of the attribute to retrieve.

    Returns
    -------
    Tuple[bool, str]
        ``(True, value)`` if the value exists and is not empty, otherwise ``(False, '')``.
    """
    att_value = note.get_ann_att_value_by_name(element) if note else None
    value = att_value.value if att_value else ''
    return value != '' and value != '-', value


class AnnotationRule(Rule):
    """Abstraction for rules that have a note type as parameter."""

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the note type parameter."""
        # backward compatibility
        parameter = parameter.replace('notetype=', '-t ') if parameter else ''
        parser = ParameterParser(prog='')
        parser.add_argument('-t', '--note_type', metavar='<type>', help='note type', required=True)
        self.add_arguments(parser)
        # parse the parameter and store the result for derived classes
        self.options = parser.parse_command(parameter)
        if self.options:
            # search for the note type
            self.note_type = get_note_type(model, self.options.note_type)
            if not self.note_type:
                message = f"'{self.options.note_type}': unknown note type"
            else:
                return Rule.OK
        else:
            message = parser.message

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def add_arguments(self, parser: ArgumentParser):
        """Declare arguments in addition to the note type."""
        pass



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ForbiddenKeywords rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from pathlib import Path

import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class ForbiddenKeywords(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0022',
        category='Naming',
        severity=Rule.MANDATORY,
        types=None,
        parameter='keywords.txt',
        description=(
            'Specific keywords shall not be used as name.\n'
            '\n'
            'The parameter is a file containing the keywords, one word per line.'
        ),
        label='Forbidden Keyword',
    ):
        if not types:
            types = [
                suite.ConstVar,
                suite.Assertion,
                suite.NamedType,
                suite.Operator,
                suite.Diagram,
                suite.ControlBlock,
                suite.Package,
                suite.State,
                suite.Label,
                suite.CompositeElement,
            ]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=types,
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        path = Path(model.descriptor.model_file_name).parent / parameter
        try:
            file = path.open(encoding='utf-8')
            lines = file.read().split('\n')
        except BaseException as e:
            self.set_message(str(e))
            # scade is a CPython module defined dynamically
            scade.output(str(e) + '\n')  # type: ignore
            return Rule.ERROR
        else:
            file.close()

        self.keywords = {_ for _ in lines if _ and _[0] != '#'}
        return Rule.OK

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # types can be redefined: protect against wrong input
        try:
            name = object.name
        except AttributeError:
            return Rule.NA

        if name in self.keywords:
            self.set_message('Forbidden keyword used as name')
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ForbiddenKeywords()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the AllInEqSet rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class AllInEqSet(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0093',
        label='The element belong to at least one equation set',
        description=(
            'The elements, except actions, assertions, and control blocks, shall belong '
            'to at least one equation set, except following use case:\n'
            '* The element is text: textual diagram, textual scope'
        ),
        category='Traceability',
        severity=Rule.ADVISORY,
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[
                suite.Equation,
                suite.WhenBranch,
                suite.IfNode,
                suite.State,
                suite.MainTransition,
                suite.ForkedTransition,
            ],
            has_parameter=False,
            **kwargs,
        )

    def on_check(self, presentable: suite.Presentable, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        pe = presentable.presentation_element
        if not pe or isinstance(pe.diagram, suite.TextDiagram):
            status = Rule.OK
        else:
            if presentable.equation_sets:
                status = Rule.OK
            else:
                status = Rule.FAILED
                message = 'the element {} shall belong to at least one equation set'
                self.set_message(message.format(presentable.get_full_path()))
        return status


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    AllInEqSet()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoStructTableComparisons rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.apitools.expr import Eck
from ansys.scade.apitools.query import get_leaf_type, is_array, is_structure
from ansys.scade.design_rules.utils.rule import Rule


class NoStructTableComparisons(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0062',
        category='Modelling',
        severity=Rule.REQUIRED,
        description='Structures and/or tables should not be compared directly with each other.',
        label='No structure or table comparisons',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        comparison_found = False
        self.warning = 0
        self.warning_text = ''
        predef_op = object_.predef_opr

        if Eck(predef_op) in {Eck.LESS, Eck.LEQUAL, Eck.GREAT, Eck.GEQUAL, Eck.EQUAL, Eck.NEQUAL}:
            for operand in object_.parameters:
                comparison_found = self._is_expression_struct_or_array(operand)
                if comparison_found:
                    break

        if comparison_found:
            container = self.get_closest_annotatable(object_)
            identifier = object_.to_string()
            error_msg = f'Comparison of structures/arrays found ( {identifier} ) '
            self.add_rule_status(container, Rule.FAILED, error_msg, identifier)
        elif self.warning == 2:
            # 2 warnings means none of the parameters could be assesed
            container = self.get_closest_annotatable(object_)
            error_msg = f'Check incomplete: {self.warning_text}'
            identifier = object_.to_string()
            self.add_rule_status(container, Rule.FAILED, error_msg, identifier)

        return Rule.NA

    def _is_expression_struct_or_array(self, operand: suite.Expression) -> bool:
        if isinstance(operand, suite.ExprId):
            reference = operand.reference
            if reference:
                return self._is_type_struct_or_array(reference.type)
            # else: incomplete model
        elif isinstance(operand, suite.ExprCall):
            operator = operand.operator
            if operator:
                # user defined operator: must have a single output to be used in a comparison
                if operand.modifier:
                    # partial iterators (w - wi) are not considered: more than one output
                    code = Eck(operand.modifier.predef_opr) if operand.modifier else Eck.NONE
                    if code in {Eck.MAP, Eck.MAPI}:
                        # map produces an array
                        return True
                    elif code in {Eck.MAPFOLD, Eck.MAPFOLDI}:
                        # either the iterated operator's output if 1 accumulator, otherwise True
                        count = int(operand.modifier.parameters[1].value)
                        if count == 0:
                            # one may wonder why using mapfold without accumulator
                            return True
                    elif code in {Eck.FOLD, Eck.FOLDI}:
                        # fall through: the output is the operator's output
                        pass
                # default: regular calls or fold[i] iterations
                # assert len(operator.outputs) == 1
                return self._is_type_struct_or_array(operator.outputs[0].type)
            else:
                code = Eck(operand.predef_opr)
                if code in {
                    Eck.CHANGE_ITH,
                    Eck.BLD_STRUCT,
                    Eck.SCALAR_TO_VECTOR,
                    Eck.BLD_VECTOR,
                    Eck.MAKE,
                    Eck.REVERSE,
                    Eck.TRANSPOSE,
                    Eck.SLICE,
                    Eck.CONCAT,
                }:
                    # these operators produce an array or a structure
                    return True
                elif code == Eck.PRJ:
                    flow = operand.parameters[0]
                    if not isinstance(flow, suite.ExprId):
                        # the projections not applied to a variable are not considered:
                        # algorithm too complex for an unlikely use case
                        self.warning += 1
                        self.warning_text = 'check for textual projections missing'
                        return False
                    if flow.reference and flow.reference.type:
                        type_ = flow.reference.type
                        for path in operand.parameters[1:]:
                            type_ = get_leaf_type(type_)
                            # protection against incorrect models
                            try:
                                if isinstance(path, suite.ConstValue) and path.kind == 'Label':
                                    name = path.value
                                    type_ = next(_.type for _ in type_.elements if _.name == name)
                                else:
                                    # get the base type of the array
                                    type_ = type_.type
                            except BaseException:
                                # more to address before involving this rule
                                return False
                        return self._is_type_struct_or_array(type_)
                elif code == Eck.PRJ_DYN:
                    # recurse with the default value
                    return self._is_expression_struct_or_array(operand.parameters[-1])
                # else: iterated predefined operators not considered

        # other use cases: scalar type
        return False

    def _is_type_struct_or_array(self, type_: suite.Type) -> bool:
        return is_array(type_) or is_structure(type_)


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoStructTableComparisons()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoLiterals rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

import ansys.scade.design_rules.utils.predef as predef
from ansys.scade.design_rules.utils.rule import Rule


class NoLiterals(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0058',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='true,false',
        label='No literals',
        description=(
            'Literals shall only be used in constant values.\n'
            "parameter: list of exceptions separated by comma: e.g.: 'true,false'"
        ),
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.ConstValue],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        if parameter is None:
            return Rule.ERROR
        self.exceptions = [_.strip() for _ in parameter.split(',')]

        return Rule.OK

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """
        Return the evaluation status for the input object.

        This rule does not apply for:
        * constants
        * type of internal variables
        * number of accumulators for iterators
        * patterns of case operator
        * labels of projections
        """
        if object_.kind == 'Label':
            # projection
            return Rule.NA
        elif self._is_case_pattern(object_):
            return Rule.NA
        elif self._is_accumulator_number(object_):
            return Rule.NA

        container = self.get_closest_annotatable(object_)
        if isinstance(container, suite.Constant):
            return Rule.NA
        elif isinstance(container, suite.LocalVariable) and container.is_internal():
            # literals accepted for computed types of internal variables
            return Rule.NA

        violated = object_.value not in self.exceptions
        if violated:
            error_msg = f'Literal found outside constant value ({object_.value})'
            identifier = object_.value
            self.add_rule_status(container, Rule.FAILED, error_msg, identifier)

        return Rule.NA

    def _is_case_pattern(self, expression: suite.Expression) -> bool:
        sequence = expression.owner
        if (
            not isinstance(sequence, suite.ExprCall)
            or sequence.predef_opr != predef.SC_ECK_SEQ_EXPR
        ):
            return False
        call = sequence.owner
        if not isinstance(call, suite.ExprCall) or call.predef_opr != predef.SC_ECK_CASE:
            return False
        return sequence == call.parameters[2]

    def _is_accumulator_number(self, expression: suite.Expression) -> bool:
        call = expression.owner
        mapfolds = {
            predef.SC_ECK_MAPFOLD,
            predef.SC_ECK_MAPFOLDI,
            predef.SC_ECK_MAPFOLDW,
            predef.SC_ECK_MAPFOLDWI,
        }
        return (
            isinstance(call, suite.ExprCall)
            and call.predef_opr in mapfolds
            and expression == call.parameters[1]
        )


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoLiterals()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the EnumDefinitionElementsOrder rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.modelling import get_enum_value
from ansys.scade.design_rules.utils.rule import Rule


def is_asc(values: list):
    """Return whether the values are in ascending order."""
    return all(values[i + 1] > values[i] for i in range(len(values) - 1))


def is_desc(values: list):
    """Return whether the values are in descending order."""
    return all(values[i + 1] < values[i] for i in range(len(values) - 1))


class EnumDefinitionElementsOrder(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0016',
        category='Types',
        label='Enumeration definition elements order',
        severity=Rule.ADVISORY,
        parameter='order=asc,by=name',
        description=(
            'Checks if the order of definition elements of enumerations is '
            'ascending (``order=asc``) or descending (``order=desc``).\n'
            'The elements are ordered by name (``by=name``) or by value '
            '(``by=value``).'
        ),
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Enumeration],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        # whether all enumeration values must have a kcg pragma value
        self.strict = False

        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            order = d.get('order')
            by = d.get('by')
            if not order:
                message = f"'{parameter}': missing 'order' value"
            elif not by:
                message = f"'{parameter}': missing 'by' value"
            else:
                self.order = order
                self.by = by
                return Rule.OK

        self.set_message(message)
        return Rule.ERROR

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if self.by == 'name':
            values = [_.name for _ in object.values]
        else:
            values = []
            last_value = -1
            missing_enum_values = []
            for value in object.values:
                v = get_enum_value(value)
                if v == -1:
                    missing_enum_values.append(value.name)
                    # C, C++, C#... semantic
                    v = last_value + 1
                last_value = v
                values.append(v)

            if self.strict and len(missing_enum_values) > 0:
                self.set_message(
                    f'Missing Enum values. Cannot check for order: {", ".join(missing_enum_values)}'
                )
                return Rule.FAILED

        if self.order == 'asc' and not is_asc(values):
            self.set_message('Value list is not in ascending order')
            return Rule.FAILED
        elif self.order == 'desc' and not is_desc(values):
            self.set_message('Value list is not in descending order')
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    EnumDefinitionElementsOrder()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ScadeVersion rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from pathlib import Path
import re

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class ScadeVersion(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0080',
        category='Tool',
        severity=Rule.REQUIRED,
        description=(
            'This rule checks if the version of the SCADE Suite models '
            'corresponds to a given version of SCADE.\n'
            "The parameter defines the version, for example '242' for SCADE Suite 2024 R2."
        ),
        label='Check the version of SCADE Suite models',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param='242',
            description=description,
            label=label,
            types=[suite.StorageUnit],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        versions = [
            # 2019 R2
            (194, {'xmlns': 6, 'xmlns:ed': 6, 'xmlns:kcg': 3}),
            # 2020 R1
            (201, {'xmlns': 6, 'xmlns:ed': 7, 'xmlns:kcg': 3}),
            # 2022 R2
            (222, {'xmlns': 6, 'xmlns:ed': 8, 'xmlns:kcg': 3}),
        ]

        if not parameter.isdecimal():
            message = '{}: Incorrect version. The parameter must be an integer, for example: 242'
            self.set_message(message.format(parameter))
            return Rule.ERROR
        version = int(parameter)
        if version < versions[0][0]:
            self.set_message(f'{parameter}: Versions prior to 2019 R2 (194) are not supported.')
            return Rule.ERROR

        # get the namespaces corresponding to the checked version
        for v in versions:
            if v[0] <= version:
                self.namespaces = v[1]
        return Rule.OK

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        path = Path(object_.sao_file_name)
        try:
            with path.open('r') as f:
                line = f.readline()
                if line.startswith('<?xml'):
                    line = f.readline()
                ns = re.findall(r'(xmlns[^=]*)="([^"]*)"', line)
                version = {_[0]: int(_[1].split('/')[-1]) for _ in ns}
        except (IOError, OSError):
            self.set_message('{0}: I/O Error'.format(path))
            return Rule.FAILED

        if version != self.namespaces:
            # report the first storage element contained in the file
            element = next((_ for _ in object_.elements if not isinstance(_, suite.Open)), None)
            if element:
                message = f'SCADE Version in xscade files does not match {parameter}'
                self.add_rule_status(element, Rule.FAILED, message)
                return Rule.NA
            # otherwise, report the storage unit itself
            self.set_message(f'{path}: SCADE Version does not match {parameter}')
            return Rule.FAILED

        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ScadeVersion()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NameStructurePackage rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.naming import is_pascal_case
from ansys.scade.design_rules.utils.rule import SCK, Rule


class NameStructurePackage(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0052',
        label='Package name',
        description='Package names shall be short (at most 10 characters for example).',
        category='Naming',
        severity=Rule.REQUIRED,
        parameter='10',
        **kwargs,
    ):
        # default value for kinds when neither types nor kinds is specified
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            kinds=[SCK.PACKAGE],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        name = object_.name
        length = int(parameter) if parameter else 0
        if len(name) > length:
            message = f'{name}: The name is longer than {length:.0f}'
            self.set_message(message)
            status = Rule.FAILED
        elif not is_pascal_case(name):
            status = Rule.FAILED
            message = f'{name}: The name shall be composed of a sequence of words'
            self.set_message(message)
        else:
            status = Rule.OK
        return status


if __name__ == '__main__':  # pragma: no cover
    NameStructurePackage()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LLRNatureEqs rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.traceability.llr_nature import LLRNature

# ignore F401: Rule required for the documentation process
from ansys.scade.design_rules.utils.rule import Rule  # noqa: F401


class LLRNatureEqs(LLRNature):
    """Implements the rule interface."""

    def __init__(self, id='id_0095', **kwargs):
        super().__init__(id=id, types=[suite.EquationSet, suite.TextDiagram], **kwargs)


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LLRNatureEqs()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NumberOfNestedSMs metric."""

if __name__ == '__main__':  # pragma: no cover
    # metric instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.metric import Metric


class NumberOfNestedSMs(Metric):
    """Implements the metric interface."""

    def __init__(
        self,
        id='id_0129',
        category='Counters',
        label='Number of nested state machines',
        description='Number of nested state machines.',
    ):
        super().__init__(
            id=id,
            label=label,
            category=category,
            description=description,
            types=[suite.StateMachine],
            kinds=None,
        )

    def on_compute_ex(self, object_: suite.Object) -> int:
        """Compute the metric for the input object."""
        self.nested_sms = 0
        self._check_sm(object_, 1)
        self.set_result_metric(self.nested_sms)

        return Metric.OK

    def _check_sm(self, state_machine: suite.StateMachine, level: int):
        if level > self.nested_sms:
            self.nested_sms = level
        for state in state_machine.states:
            self._check_action_or_state(state, level)

    def _check_activate_block(self, activate_block: suite.ActivateBlock, level: int):
        if isinstance(activate_block, suite.IfBlock):
            self._check_if_branch(activate_block.if_node, level)
        else:
            # assert isinstance(activate_block, suite.WhenBlock):
            for when_branch in activate_block.when_branches:
                self._check_action_or_state(when_branch.action, level)

    def _check_if_branch(self, if_branch: suite.IfBranch, level: int):
        if isinstance(if_branch, suite.IfNode):
            self._check_if_branch(if_branch.then, level)
            self._check_if_branch(if_branch._else, level)
        else:
            # assert isinstance(if_branch, suite.IfAction):
            self._check_action_or_state(if_branch.action, level)

    def _check_action_or_state(self, datadef: suite.DataDef, level: int):
        for activate_block in datadef.activate_blocks:
            self._check_activate_block(activate_block, level)
        for state_machine in datadef.state_machines:
            self._check_sm(state_machine, level + 1)


if __name__ == '__main__':  # pragma: no cover
    # metric instantiated outside of a package
    NumberOfNestedSMs()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NameStructureDiagram rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import re

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import SCK, Rule


class NameStructureDiagram(Rule):
    """Implements the rule interface."""

    pattern_multiple = re.compile(r'(.*[^0-9])\d+')

    def __init__(
        self,
        id='id_0047',
        label='Diagram name',
        description=(
            'When there are several diagrams in a scope, '
            'the name of each diagram shall be characteristic of its function.\n'
            'Otherwise, the default name created by the editor shall be '
            'updated to the name of its scope.'
        ),
        category='Naming',
        severity=Rule.ADVISORY,
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            kinds=[SCK.DIAGRAM],
            has_parameter=False,
            **kwargs,
        )

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        name = object_.name
        scope = object_.owner
        scope_name = '' if isinstance(scope, suite.Action) else scope.name
        if len(scope.diagrams) == 1:
            if not scope_name:
                # TODO(Jean): one may want top compare the name with the one
                #             we get from get_full_path
                # https://github.com/ansys/scade-design-rules/issues/29
                status = Rule.NA
            elif name != scope.name:
                # TODO(Jean): As an alternative, make sure the name of the scope is
                #             included in the name of the diagram
                # https://github.com/ansys/scade-design-rules/issues/29
                message = f'{name}: The name shall be the name of its scope {scope.name}'
                self.set_message(message)
                status = Rule.FAILED
            else:
                status = Rule.OK
        else:
            # fail when the name differ from a sibling by a suffix number
            # or that has the same name as the scope
            pattern_scope = re.compile(scope_name + r'_?\d+')
            if pattern_scope.fullmatch(name):
                # check of diagram name vs owner name
                message = f"{name}: The name derives from its scope's name instead of a description"
                self.set_message(message)
                status = Rule.FAILED
            else:
                match = self.pattern_multiple.fullmatch(name)
                if match:
                    # check for a similar sibling name => copies
                    prefix = match.group(1)
                    pattern_prefix = re.compile(prefix + r'\d+')
                    for diagram in scope.diagrams:
                        if diagram != object_ and pattern_prefix.fullmatch(diagram.name):
                            message = f'{name}: The name shall be a description'
                            self.set_message(message)
                            status = Rule.FAILED
                            break
                    else:
                        status = Rule.OK
                else:
                    status = Rule.OK
        return status


if __name__ == '__main__':  # pragma: no cover
    NameStructureDiagram()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NameStructureIteratorIndex rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import re

import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.command import ParameterParser
from ansys.scade.design_rules.utils.modelling import IR, get_iter_role
from ansys.scade.design_rules.utils.rule import SCK, Rule


class NameStructureIteratorIndex(Rule):
    """Implements the rule interface."""

    # parameter = "index = index"
    def __init__(
        self,
        id='id_0051',
        label='Name structure of iterator index',
        description=(
            'For iterators which use an index (mapi, mapwi, foldi, foldwi, '
            'mapfoldi, mapfoldwi), the input corresponding to the index shall be named'
            " or prefixed by 'index'."
        ),
        category='Naming',
        severity=Rule.REQUIRED,
        parameter='-i index',
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            kinds=[SCK.INPUT, SCK.HIDDEN],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )
        self.index_regexp = ''

    def on_start(self, model: suite.Model, parameter: str) -> int:
        """Get the rule's parameters."""
        parameter = parameter.replace('index=', '-i ') if parameter else ''
        parser = ParameterParser(prog='')
        parser.add_argument('-i', '--index', help='Regular expression for index', required=True)
        options = parser.parse_command(parameter)
        if not options:
            message = parser.message
        else:
            self.index_regexp = options.index
            return Rule.OK

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check_ex(self, variable: suite.LocalVariable, parameter: str = '') -> int:
        """
        Return the evaluation status for the input object.

        Retrieve the variable's roles for all instances of its operator, fail when:

        * the variable is used as index and does not match the index
          regular expression
        """
        failure = False
        name = variable.name

        # collect all roles in a set to avoid redundant failures
        roles = {get_iter_role(variable, call) for call in variable.operator.expr_calls}
        lines = []
        for role in roles:
            if role == IR.INDEX:
                if not re.fullmatch(self.index_regexp, name):
                    failure = True
                    lines.append(
                        f'The name does not match the index expression {self.index_regexp}'
                    )

        if failure:
            self.set_message(',\n'.join(lines))
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NameStructureIteratorIndex()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Implements the CheckConfigurationCustom1 rule.

This rule has been developed specifically for a customer needs.
It may fulfill very specific purpose.
If you need to change this rule, consider forking/deriving to a new rule.
"""

if __name__ == '__main__':
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class CheckConfigurationCustom1(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0104',
        category='configuration',
        severity=Rule.MANDATORY,
        parameter='conf=KCG,rootPackage=package1',
        description=('Check the given configuration using specific parameters.\n'),
        label='Check configuration',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Model],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            conf = d.get('conf')
            root_package = d.get('rootPackage')
            if not conf:
                message = f"'{parameter}': missing 'conf' value"
            elif not root_package:
                message = f"'{parameter}': missing 'rootPackage' value"
            else:
                self.conf = conf
                self.root_package = root_package
                return Rule.OK

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        failure_messages = []

        assert isinstance(object_, suite.Model)  # nosec B101  # addresses linter
        project = object_.project
        # check if object (model) is the main project and not a library model
        if not project:
            return Rule.OK

        configuration = project.find_configuration(self.conf)

        # 1 check for "no skip unused model objects"
        if project.get_bool_tool_prop_def('GENERATOR', 'SKIP_UNUSED', False, configuration):
            failure_messages.append('Skip unused model object is set')

        # 2 check for root operator setting
        rootnode_start = 'SoftwareComponents::ACD_Appl::'
        values = project.get_tool_prop_def('GENERATOR', 'ROOTNODE', [], configuration)
        for prop_value in values:
            if prop_value.startswith(rootnode_start):
                break
        else:
            failure_messages.append(f'Root node does not start with: {rootnode_start}')

        # 3 Code Integration settings
        if project.get_scalar_tool_prop_def('GENERATOR', 'TARGET_ADAPTOR', '', configuration):
            failure_messages.append('Code Integration Settings not correct')

        # 4 Optimization Settings
        if project.get_bool_tool_prop_def('GENERATOR', 'STATIC', False, configuration):
            failure_messages.append('Local variables as static is set')
        if project.get_bool_tool_prop_def('GENERATOR', 'INPUT_THRESHOLD', False, configuration):
            failure_messages.append('Input Threshold is set')

        # 5 Check User Config Setting
        user_config = '..\\..\\include\\' + object_.name + '_user_macros.h'
        value = project.get_scalar_tool_prop_def('GENERATOR', 'USER_CONFIG', '', configuration)
        if value != user_config:
            failure_messages.append(f'User Config not set to: {user_config}')

        # 6 Compiler settings
        value = project.get_scalar_tool_prop_def('SIMULATOR', 'CPU_TYPE', '', configuration)
        if value != 'win64':
            failure_messages.append('Simulator CPU type not correct')

        # 7 check annotation note MyAnnotationType_suite
        descriptor = object_.descriptor
        aty = (
            '..\\..\\..\\..\\architecture\\scade_plugins\\'
            'fcmsAnnotationType\\MyAnnotationType_suite.aty'
        )
        if aty not in descriptor.ann_type_files:
            failure_messages.append('MyAnnotationType_suite.aty not set properly')

        # 7 Prefix settings
        number_of_public = 0
        prefix_desired = ''
        root_package = None
        # assumption: top level package is "SoftwareComponents::ACD_Appl::"
        for package in object_.all_packages:
            if package.name == self.root_package:
                root_package = package
                break
        if root_package is not None:
            # assumption: only one public operator which is the root operator
            for operator in root_package.operators:
                if operator.visibility == 'Public':
                    number_of_public += 1
                    prefix_desired = (operator.name.split('_'))[-1]
        value = project.get_scalar_tool_prop_def('GENERATOR', 'GLOBALS_PREFIX', '', configuration)
        if value:
            if number_of_public == 1 and value != prefix_desired:
                failure_messages.append(f'Global Prefix not set to: {prefix_desired}')
        else:
            failure_messages.append('Global Prefix not set')

        if failure_messages:
            self.set_message(
                f'Configuration not correct ({len(failure_messages)}): '
                + ', '.join(failure_messages)
            )
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':
    # rule instantiated outside of a package
    CheckConfigurationCustom1()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utilities for developing naming rules."""

import re
from typing import Dict, List

# list from Scade LRM, section 2.4, KCG 6.6
_keywords = set(
    """
abstract, activate, and, assume, automaton,
bool,
case, char, clock, const,
default, do,
else, elsif, emit, end, enum, every,
false, fby, final, flatten, float, float32, float64, fold, foldi, foldw, foldwi, function,
guarantee, group,
if, imported, initial, int8, int16, int32, int64, integer, is,
land, last, let, lnot, lor, lsl, lsr, lxor,
make, map, mapfold, mapfoldi, mapfoldw, mapfoldwi, mapi, mapw, mapwi, match, merge, mod,
node, not, numeric,
of, onreset, open, or,
package, parameter, pre, private, probe, public,
restart, resume, returns, reverse,
sensor, sig, signed, specialize, state, synchro,
tel, then, times, transpose, true, type,
uint8, uint16, uint32, uint64, unless, unsigned, until,
var,
when, where, with,
xor
""".replace('\n', ' ')
    .replace(',', '')
    .split()
)


def is_scade_keyword(name: str) -> bool:
    """
    Return whether the name is a Scade keyword.

    Parameters
    ----------
    name : str
        Input identifier.

    Returns
    -------
    bool
    """
    return name in _keywords


# the patterns accept acronyms of 2 letters at most
# cf. https://docs.microsoft.com/en-us/previous-versions/dotnet/netframework-1.1/141e06ef(v=vs.71)/
_pascal_case_pattern = re.compile('(?:[A-Z]{1,3}[a-z0-9]+)*(?:[A-Z]{1,2})?')
_camel_case_pattern = re.compile('[a-z0-9]+(?:[A-Z]{1,3}[a-z0-9]+)*(?:[A-Z]{1,2})?')


def is_camel_case(name: str) -> bool:
    """
    Return whether the name matches the camel case pattern.

    Parameters
    ----------
    name : str
        Input identifier.

    Returns
    -------
    bool
    """
    return _camel_case_pattern.fullmatch(name) is not None


def is_pascal_case(name: str) -> bool:
    """
    Return whether the name matches the pascal case pattern.

    Parameters
    ----------
    name : str
        Input identifier.

    Returns
    -------
    bool
    """
    return _pascal_case_pattern.fullmatch(name) is not None


def substitute_names(text: str, aliases: Dict[str, str]) -> str:
    """
    Return a text with substitutions of words.

    Parameters
    ----------
    text : str
        Input text.
    aliases : Dict[str, str]
        Dictionary of replacements.

    Returns
    -------
    str
        Updated text.
    """
    for alias, value in aliases.items():
        text = re.sub(rf'\b{alias}\b', value, text)
    return text


def tokenize_name(word: str) -> List[str]:
    """
    Return the list of tokens constituting a word.

    Parameters
    ----------
    word : str
        Input word.

    Returns
    -------
    List[str]
        List of tokens.
    """
    tokens = []
    for name in word.split('_'):
        if name == '':
            # word is prefixed or suffixed by '_'
            tokens.append('')
            continue
        # there must be a more clever algorithm, one pass
        # for now, adapt an existing one
        prev = name[0]
        lower = prev
        for c in name[1:]:
            if c.isupper() and prev != '_':
                lower += '_'
            prev = c
            lower += c
        lst = lower.split('_')
        # merge successive uppercase singletons
        prev = lst[0]
        for token in lst[1:]:
            if prev.isupper() and len(token) == 1 and token.isupper():
                prev += token
            else:
                tokens.append(prev)
                prev = token
        tokens.append(prev)
    return tokens



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the AllConstantsTypesAreUsed rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.annotations import (
    AnnotationRule,
    get_first_note_by_type,
    is_ann_note_value_defined_and_get_value,
)
from ansys.scade.design_rules.utils.rule import Rule


class AllConstantsTypesAreUsed(AnnotationRule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0002',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='-t ConstantUsage',
        label='All constants and types are used at least once',
        types=None,
        description=(
            'All constants and types are used at least once.\nFor constants, it is also checked '
            "if the annotation note 'only external use' is set.\n"
            'parameter:\n'
            "* '-t': Name of the annotation note type\n"
        ),
    ):
        if not types:
            types = [suite.Constant, suite.NamedType]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=types,
            kinds=None,
        )

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violated = False

        if isinstance(object, suite.NamedType):
            if not object.typed_objects:
                violated = True
        elif isinstance(object, suite.Constant):
            if not object.expr_ids:
                # check for annotation note for external use mark
                ann_note = get_first_note_by_type(object, self.note_type)
                defined, value = is_ann_note_value_defined_and_get_value(
                    ann_note, 'isExternallyUsed_TrueFalse'
                )
                violated = not defined or not value
        else:
            # scade is a CPython module defined dynamically
            scade.output(f'Rule not implemented for {object.__class__.__name__}')  # type: ignore
            Rule.ERROR

        if violated:
            self.set_message(f'{object.__class__.__name__} {object.name} not used in model.')
            return Rule.FAILED

        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    AllConstantsTypesAreUsed()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NameStructureFoldAccumulator rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import re

import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.command import ParameterParser
from ansys.scade.design_rules.utils.modelling import IR, get_iter_role
from ansys.scade.design_rules.utils.rule import SCK, Rule


class NameStructureFoldAccumulator(Rule):
    """Implements the rule interface."""

    # parameter = "in = acc.*In, out = acc.*Out, strict =False"
    def __init__(
        self,
        id='id_0049',
        label='Name structure of accumulator inputs/outputs',
        description=(
            'For the operators iterated with fold constructs (fold, foldi, foldw, '
            'foldwi, mapfold, mapfoldi, mapfoldw, mapfoldwi):\n'
            "* The accumulators are defined by a name prefixed by 'acc'.\n"
            '* The input/output variables of an accumulator shall be discriminated by '
            "  the suffix 'In' for the inputs and 'Out' for the outputs.\n"
            '* The input and output names of the accumulators match.\n'
            '\n'
            'The parameter allows specifying a regular expression for input and output names.\n'
            "For example: '-i acc(.*)In -o acc(.*)Out'. The parentheses identify the common "
            'part of both names that must match.\n'
            '\n'
            'When strict is set, the rules verifies the names are not used for variables '
            'which are not accumulators.'
        ),
        category='Naming',
        severity=Rule.REQUIRED,
        parameter='-i acc(.*)In -o acc(.*)Out',
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[],
            kinds=[SCK.INPUT, SCK.HIDDEN, SCK.OUTPUT],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )
        self.in_regexp = ''
        self.out_regexp = ''
        self.strict = False

    def on_start(self, model: suite.Model, parameter: str) -> int:
        """Get the rule's parameters."""
        # minimal level of backward compatibility
        parameter = (
            parameter.replace('in=', '-i ').replace(',', ' ').replace('out=', '-o ')
            if parameter
            else ''
        )
        parameter = (
            parameter.replace('strict=true', '-s ').replace('strict=false', '') if parameter else ''
        )
        parser = ParameterParser(prog='')
        help_in = 'Regular expression for the name of the accumulator inputs'
        parser.add_argument(
            '-i', '--in', dest='in_', help=help_in, required=True, metavar='<regular expression>'
        )
        help_out = 'Regular expression for the name of the accumulator outputs'
        parser.add_argument(
            '-o', '--out', help=help_out, required=True, metavar='<regular expression>'
        )
        help_strict = (
            'Optional parameter to prevent the usage of ``in``/``out`` '
            'expressions for variables that are not accumulators'
        )
        parser.add_argument('-s', '--strict', help=help_strict, action='store_true')
        options = parser.parse_command(parameter)
        if not options:
            message = parser.message
        else:
            self.in_regexp = options.in_
            self.out_regexp = options.out
            self.strict = options.strict
            return Rule.OK

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check_ex(self, variable: suite.LocalVariable, parameter: str = '') -> int:
        """
        Return the evaluation status for the input object.

        Retrieve the variable's roles for all instances of its operator and fail when:

        * The variable is an input (resp. output) accumulator and does not
          match the in (resp. out) regular expression
        * The variable is not an accumulator and matches one of the in/out
          regular expressions
        """
        failure = False
        name = variable.name

        # collect all roles in a set to avoid redundant failures
        operator = variable.operator
        roles = {get_iter_role(variable, call) for call in operator.expr_calls}
        lines = []
        for role in roles:
            if role == IR.ACC_IN:
                m = re.fullmatch(self.in_regexp, name)
                if not m:
                    failure = True
                    lines.append(
                        f'The name does not match the input accumulator expression {self.in_regexp}'
                    )
                else:
                    groups = m.groups()
                    if groups:
                        base_name = groups[0]
                        # search the operator for a corresponding input
                        for input in operator.outputs:
                            m = re.fullmatch(self.out_regexp, input.name)
                            if m and m.groups() and m.groups()[0] == base_name:
                                # match found
                                break
                        else:
                            failure = True
                            lines.append(f'Matching output accumulator not found: {variable.name}')
            elif role == IR.ACC_OUT:
                m = re.fullmatch(self.out_regexp, name)
                if not m:
                    failure = True
                    message = 'The name does not match the output accumulator expression {}'
                    lines.append(message.format({self.out_regexp}))
                else:
                    groups = m.groups()
                    if groups:
                        base_name = groups[0]
                        # search the operator for a corresponding input
                        for input in operator.inputs + operator.hiddens:
                            m = re.fullmatch(self.in_regexp, input.name)
                            if m and m.groups() and m.groups()[0] == base_name:
                                # match found
                                break
                        else:
                            failure = True
                            lines.append(
                                f'Matching input accumulator not found for {variable.name}'
                            )
            else:
                if self.strict and (
                    re.fullmatch(self.in_regexp, name) or re.fullmatch(self.out_regexp, name)
                ):
                    failure = True
                    lines.append('The variable is not used as an accumulator')

        if failure:
            self.set_message(',\n'.join(lines))
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NameStructureFoldAccumulator()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides the Scade predefined operator codes as Python constants."""

SC_ECK_NONE = 1
SC_ECK_AND = 2
SC_ECK_OR = 3
SC_ECK_XOR = 4
SC_ECK_NOT = 5
SC_ECK_SHARP = 6
SC_ECK_PLUS = 7
SC_ECK_SUB = 8
SC_ECK_NEG = 9
SC_ECK_MUL = 10
SC_ECK_REAL2INT = 11
SC_ECK_INT2REAL = 12
SC_ECK_SLASH = 14
SC_ECK_DIV = 15
SC_ECK_MOD = 16
SC_ECK_PRJ = 18
SC_ECK_CHANGE_ITH = 19
SC_ECK_LESS = 20
SC_ECK_LEQUAL = 21
SC_ECK_GREAT = 22
SC_ECK_GEQUAL = 23
SC_ECK_EQUAL = 24
SC_ECK_NEQUAL = 25
SC_ECK_PRE = 26
SC_ECK_WHEN = 28
SC_ECK_FOLLOW = 29
SC_ECK_FBY = 30
SC_ECK_IF = 31
SC_ECK_CASE = 32
SC_ECK_SEQ_EXPR = 33
SC_ECK_BLD_STRUCT = 34
SC_ECK_MAP = 35
SC_ECK_FOLD = 36
SC_ECK_MAPFOLD = 37
SC_ECK_MAPI = 38
SC_ECK_FOLDI = 39
SC_ECK_SCALAR_TO_VECTOR = 40
SC_ECK_BLD_VECTOR = 41
SC_ECK_PRJ_DYN = 42
SC_ECK_MAKE = 43
SC_ECK_FLATTEN = 44
SC_ECK_MERGE = 45
SC_ECK_REVERSE = 46
SC_ECK_TRANSPOSE = 47
SC_ECK_TIMES = 49
SC_ECK_MATCH = 50
SC_ECK_SLICE = 51
SC_ECK_CONCAT = 52
SC_ECK_ACTIVATE = 53
SC_ECK_RESTART = 54
SC_ECK_FOLDW = 55
SC_ECK_FOLDWI = 56
SC_ECK_ACTIVATE_NOINIT = 57
SC_ECK_CLOCKED_ACTIVATE = 58
SC_ECK_CLOCKED_NOT = 59
SC_ECK_POS = 60
SC_ECK_MAPW = 61
SC_ECK_MAPWI = 62
SC_ECK_NUMERIC_CAST = 63
SC_ECK_MAPFOLDI = 64
SC_ECK_MAPFOLDW = 65
SC_ECK_MAPFOLDWI = 66
SC_ECK_LAND = 67
SC_ECK_LOR = 68
SC_ECK_LXOR = 69
SC_ECK_LNOT = 70
SC_ECK_LSL = 71
SC_ECK_LSR = 72



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoScadeLibrary rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import os
from pathlib import Path

import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class NoScadeLibrary(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0061',
        label='SCADE libraries shall not be used in production',
        description=(
            'Only SCADE libraries that are under full control of the project'
            '(configuration management, verification) shall be used.\n'
            'In particular, the designer shall NOT use the SCADE product installation libraries, '
            'as they are delivered as examples.'
        ),
        category='Project Structure',
        severity=Rule.REQUIRED,
        parameter='upper_levels=1',
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[suite.Model],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )
        self.upper_levels = 0

    def on_start(self, model: suite.Model, parameter: str) -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            upper_levels = d.get('upper_levels')
            self.upper_levels = int(upper_levels) if upper_levels is not None else 0
            return Rule.OK

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        assert isinstance(object_, suite.Model)  # nosec B101  # addresses linter
        external_libraries = {}
        if object_.project:
            # $(SCADE) can be tested only for the main model's libraries
            for file in object_.project.file_refs:
                pathname = file.persist_as
                suffix = Path(pathname).suffix.lower()
                if suffix != '.etp' and suffix == '.vsp':
                    # not a library project
                    continue
                if '$(SCADE)' in pathname:
                    external_libraries[Path(file.pathname)] = pathname
                # TODO (Jean): this check can be optional
                # https://github.com/ansys/scade-design-rules/issues/29
                # elif (Path(pathname).is_absolute()):
                #     external_libraries.append(pathname)

        # make sure the dependencies are in sibling folders
        # or not absolute pathnames
        path_model = Path(object_.descriptor.model_file_name)
        for library in object_.libraries:
            path = Path(library.descriptor.model_file_name)
            try:
                rel_path = os.path.relpath(path, path_model.parent)
            except ValueError:  # More specific exception than BaseException
                external_libraries.setdefault(path, str(path))
                continue
            count = rel_path.count('..')
            if count > self.upper_levels and path not in external_libraries:
                external_libraries[path] = str(path)

        if external_libraries:
            status = Rule.FAILED
            message = (
                'The model {} shall not use SCADE product libraries or libraries outside '
                'of the CM workspace:\n{}'
            )
            self.set_message(message.format(object_.name, '\n'.join(external_libraries.values())))
        else:
            status = Rule.OK
        return status


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoScadeLibrary()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the CheckerNoWarningsErrors rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import re
import subprocess  # nosec B404  # used to call scade.exe -check

import scade.model.suite as suite

from ansys.scade.apitools.info.install import get_scade_home
from ansys.scade.design_rules.utils.rule import SCK, Rule


class CheckerNoWarningsErrors(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0118',
        category='Tool',
        severity=Rule.REQUIRED,
        parameter='conf=KCG',
        description=(
            'This rule checks if the semantic checker reports no errors and no warnings.\n'
            'parameter: conf=configuration name'
        ),
        label='Semantic checker shall raise no errors and warnings',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=None,
            kinds=[SCK.MODEL],
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            d = f"'{parameter}': parameter syntax error"
        else:
            configuration = d.get('conf')
            if not configuration:
                message = f"'{parameter}': missing 'conf' value"
            else:
                self.configuration = configuration
                return Rule.OK

        self.set_message(message)
        return Rule.ERROR

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violated = False

        # check if object (model) is the main project and not a library model
        if object_.library:
            return Rule.NA

        # create command line call
        scade_home = str(get_scade_home())
        scade_path = scade_home + r'\SCADE\bin\scade.exe'
        pathname = object_.descriptor.model_file_name
        command = [scade_path, '-check', pathname, '-conf', self.configuration]

        # execute call
        cp = subprocess.run(command, capture_output=True)  # nosec B603  # inputs checked
        outputs = cp.stdout.decode('utf8').split('\n')

        # parse result string for errors and warnings
        x = re.search(r'^Checker ends with (\d) error.*(\d) warning', outputs[0])
        assert x is not None  # nosec B101  # addresses linter
        errors = x.groups()[0]
        warnings = x.groups()[1]

        if int(errors) > 0 or int(warnings) > 0:
            violated = True

        if violated:
            self.set_message(outputs[0])
            return Rule.FAILED

        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    CheckerNoWarningsErrors()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LLRArchitecture rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from typing import Optional

import scade
import scade.model.suite as suite
import scade.model.suite.annotation as annot

from ansys.scade.design_rules.utils.annotations import (
    AnnotationRule,
    ArgumentParser,
    get_first_note_by_type,
)
from ansys.scade.design_rules.utils.rule import Rule


class LLRArchitecture(AnnotationRule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0035',
        label="'Architecture' CE",
        description=(
            'Architecture Contributing Elements (CE) can be only equation sets or textual diagrams.'
        ),
        category='Traceability',
        severity=Rule.ADVISORY,
        parameter='-t DesignElement -a Nature -v Architecture',
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[suite.State, suite.Transition],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )

        self.note_type: Optional[annot.AnnNoteType] = None
        self.note_attribute: Optional[annot.AnnAttDefinition] = None
        self.architecture = None

    def on_start(self, model: suite.Model, parameter: str) -> int:
        """Get the rule's parameters."""
        status = super().on_start(model, parameter)
        if status == Rule.OK:
            assert self.note_type is not None  # nosec B101  # addresses linter
            assert self.options is not None  # nosec B101  # addresses linter
            self.architecture = self.options.value
            for note_attribute in self.note_type.ann_att_definitions:
                if note_attribute.name == self.options.attribute:
                    self.note_attribute = note_attribute
                    break
            else:
                status = Rule.ERROR
                message = f"'{self.options.attribute}': unknown note attribute"
                self.set_message(message)
                # scade is a CPython module defined dynamically
                scade.output(message + '\n')  # type: ignore

        return status

    def add_arguments(self, parser: ArgumentParser):
        """Declare arguments in addition to the note type."""
        parser.add_argument(
            '-a', '--attribute', metavar='<attribute>', help='attribute name', required=True
        )
        parser.add_argument(
            '-v', '--value', metavar='<value>', help='value for architecture', required=True
        )

    def on_check(self, annotable: suite.Annotable, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        assert self.note_type is not None  # nosec B101  # addresses linter
        assert self.note_attribute is not None  # nosec B101  # addresses linter
        note = get_first_note_by_type(annotable, self.note_type)
        if note:
            for value in note.ann_att_values:
                if (
                    value.ann_att_definition == self.note_attribute
                    and value.to_string() == self.architecture
                ):
                    message = "The {} of the Contributing Element can't be '{}'"
                    self.set_message(message.format(self.note_attribute.name, self.architecture))
                    return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LLRArchitecture()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LibProjectsUsage rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class LibProjectsUsage(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0033',
        category='Project Structure',
        severity=Rule.REQUIRED,
        parameter='Lib',
        description=(
            'This rule checks if only library Projects (prefixed with Lib) are '
            'included in the current Project.'
        ),
        label="Only 'Lib' prefixed projects shall be used as library Projects",
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Model],
            kinds=None,
        )

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        return_status = Rule.OK
        return_message = []
        for library in object.libraries:
            if not library.name.startswith(parameter):
                return_message.append(f'Included Project {library.name} is no Library project')
                return_status = Rule.FAILED

        if return_status == Rule.FAILED:
            return_message.append(
                f"Library Projects shall contain the prefix '{parameter}' in their name!"
            )

        separator = '; '
        self.set_message(separator.join(return_message))
        return return_status


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LibProjectsUsage()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the CheckConfiguration rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from pathlib import Path

from defusedxml import ElementTree
import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class CheckConfiguration(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0103',
        category='configuration',
        severity=Rule.MANDATORY,
        parameter='conf=KCG,project=check.etp,conf_source=KCG',
        description='Check the given configuration against an existing project.',
        label='Check configuration',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Model],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            conf = d.get('conf')
            project = d.get('project')
            conf_source = d.get('conf_source')
            if not conf:
                message = f"'{parameter}': missing 'conf' value"
            elif not project:
                message = f"'{parameter}': missing 'project' value"
            elif not conf_source:
                message = f"'{parameter}': missing 'conf_source' value"
            else:
                self.conf = conf
                # make the path relative to the project
                path = Path(model.descriptor.model_file_name)
                self.project_source_path = path.parent / project
                self.conf_source = conf_source
                return Rule.OK

        self.set_message(message)
        return Rule.ERROR

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        assert isinstance(object_, suite.Model)  # nosec B101  # addresses linter
        project = object_.project
        configuration = project.find_configuration(self.conf)
        # load source project: nightmare
        # -> scade.load_project can't be used when the rules are evaluated from the IDE
        # -> scade.model.project.project.xml_project.load_project can't be used either
        #    when the rules are evaluated from the IDE (simple typo in the file but enough)
        # the only left option consists in parsing the xml file, with xml since lxml
        # can't be used in the IDE
        try:
            f = self.project_source_path.open()
        except IOError as e:
            self.set_message(str(e))
            return Rule.FAILED
        else:
            _tree = ElementTree.parse(f)
            f.close()

        root = _tree.getroot()

        # find configuration ID of given source configuration
        for element in root.iter():
            if element.tag == 'Configuration':
                id = element.get('id')  # configuration id
                name = element.get('name')  # configuration name
                if name == self.conf_source:
                    conf_id = id
                    break
        else:
            conf_id = -1

        props_source = {}
        # go through all Props
        for element in root.iter():
            if element.tag == 'Prop':
                values = []
                conf_found = False
                # check if configuration matches
                for sub_elem in element.iter():
                    if sub_elem.tag == 'configuration':
                        id = sub_elem.text
                        if id == conf_id:
                            conf_found = True
                    elif sub_elem.tag == 'value':
                        values.append(sub_elem.text)
                # configuration found then compare values
                if conf_found:
                    name = element.get('name')
                    props_source[name] = values

        # get conf settings of current project
        props_target = {_.name: _.values for _ in configuration.props}

        # compare values
        failure_messages = []
        set_props_target = set(props_target)
        set_props_source = set(props_source)
        diff1s = set_props_source - set_props_target
        diff2s = set_props_target - set_props_source
        for diff1 in diff1s:
            failure_messages.append(diff1)
        for diff2 in diff2s:
            if props_target[diff2] != ['']:
                failure_messages.append(diff2)
        for key, value in props_target.items():
            if key in props_source:
                source = props_source[key]
                if value != source:
                    fail_text = f'{key}({",".join(value)} <> {",".join(source)})'
                    failure_messages.append(fail_text)

        if failure_messages:
            self.set_message(f'Configuration not correct: {", ".join(failure_messages)}')
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    CheckConfiguration()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the IdenticalForProducerConsumer rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from typing import Set

import scade.model.suite as suite

from ansys.scade.design_rules.utils.annotations import (
    AnnotationRule,
    get_first_note_by_type,
    is_ann_note_value_defined_and_get_value,
)
from ansys.scade.design_rules.utils.rule import SCK, Rule


class IdenticalForProducerConsumer(AnnotationRule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0027',
        category='annotations',
        severity=Rule.REQUIRED,
        parameter='-t SDD_TopLevel',
        description=(
            'The annotation notes for producer and consumer shall be identical '
            'if they both exist.\n'
            "If only the consumer has a note raise a 'redefine' warning.\n\n"
            'parameter:\n'
            "* '-t': Name of the annotation note type"
        ),
        label='Identical annotations for producer and consumer',
        # ease customization
        min_field='Min_Value',
        max_field='Max_Value',
        unit_field='Unit_SI',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=None,
            kinds=[SCK.INPUT, SCK.OUTPUT, SCK.HIDDEN],
        )
        self.min_field = min_field
        self.max_field = max_field
        self.unit_field = unit_field

    def on_check_ex(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        self.violated = False
        self.violation_text = []

        consumer = object
        producers = set()

        if consumer.is_output():
            producers.update(self._get_producers(consumer))
        else:
            # assert consumer.is_input() or consumer.is_hidden()
            # get operator in which the inputs and hiddens are defined
            operator = consumer.operator
            # create a combined list of all inputs and hiddens.
            # This is possible since in the IDE the hiddens always come at the end
            all_inputs = operator.inputs + operator.hiddens
            # get position of input/hidden in operator interface
            input_pos_in_operator = all_inputs.index(consumer)
            # go through all occurrences where the operator is called from another operator
            for expr_call in operator.expr_calls:
                parameters = expr_call.parameters
                if len(all_inputs) - 1 == len(parameters):
                    if input_pos_in_operator == 0:
                        # it is the index of a 'i' iterator: no dependencies
                        continue
                    input_pos_in_operator -= 1
                else:
                    if len(all_inputs) != len(parameters):  # pragma: no cover
                        # incorrect model
                        continue
                param = parameters[input_pos_in_operator]
                if isinstance(param, suite.ExprId) and isinstance(
                    param.reference, suite.LocalVariable
                ):
                    producers.update(self._get_producers(param.reference))

        for producer in producers:
            self._compare_annotation_notes(producer, consumer)

        if self.violated:
            self.set_message(f'Annotation notes issues: {";".join(self.violation_text)}')
            return Rule.FAILED
        return Rule.OK

    def _get_producers(self, consumer: suite.LocalVariable) -> Set[suite.LocalVariable]:
        """Return the producers of the consumer."""
        producers = set()
        # analyze the equations in which the consumer is produced
        for prev_eq in consumer.definitions:
            right = prev_eq.right
            if isinstance(right, suite.ExprId) and isinstance(right.reference, suite.LocalVariable):
                producer = right.reference
                if producer.is_input():
                    producers.add(producer)
                else:
                    # recurse
                    producers.update(self._get_producers(producer))
            elif isinstance(right, suite.ExprCall):
                # right side is a called operator: could be inside iterator
                if right.operator:
                    # get operator which is called and find connected output.
                    called_op = right.operator
                    # get position of local variable in left side if equation.
                    local_var_pos_in_lefts = prev_eq.lefts.index(consumer)
                    if len(called_op.outputs) < len(prev_eq.lefts):
                        if local_var_pos_in_lefts == 0:
                            # exit index: no dependencies
                            continue
                        else:
                            local_var_pos_in_lefts -= 1
                    if local_var_pos_in_lefts >= len(called_op.outputs):  # pragma: no cover
                        # incorrect model
                        continue
                    producer = called_op.outputs[local_var_pos_in_lefts]
                    producers.add(producer)
                # else: predefined operator: stop the analysis
        return producers

    def _compare_annotation_notes(
        self, producer: suite.LocalVariable, consumer: suite.LocalVariable
    ):
        """Compare the annotation notes of producer and consumer."""
        self.violated = False

        ann_note_consumer = get_first_note_by_type(consumer, self.note_type)
        ann_note_producer = get_first_note_by_type(producer, self.note_type)

        # Unit SI check
        if self.unit_field:
            consumer_unit_si_defined, consumer_unit_si = is_ann_note_value_defined_and_get_value(
                ann_note_consumer, self.unit_field
            )
            producer_unit_si_defined, producer_unit_si = is_ann_note_value_defined_and_get_value(
                ann_note_producer, self.unit_field
            )
            if consumer_unit_si_defined:
                if not producer_unit_si_defined:
                    self.violated = True
                    self.violation_text.append(f'{self.unit_field}: {producer.name} is redefined.')
                elif consumer_unit_si != producer_unit_si:
                    self.violated = True
                    self.violation_text.append(
                        f'{self.unit_field}: {consumer.name} != {producer.name}.'
                    )

        # Range check min
        if self.min_field:
            consumer_min_defined, consumer_min = is_ann_note_value_defined_and_get_value(
                ann_note_consumer, self.min_field
            )
            producer_min_defined, producer_min = is_ann_note_value_defined_and_get_value(
                ann_note_producer, self.min_field
            )
            if consumer_min_defined:
                if not producer_min_defined:
                    self.violated = True
                    self.violation_text.append(f'{self.min_field}: {producer.name} is redefined.')
                elif float(consumer_min) > float(producer_min):
                    self.violated = True
                    self.violation_text.append(
                        f'{self.min_field}: {consumer.name} > {producer.name}.'
                    )

        # Range check max
        if self.max_field:
            consumer_max_defined, consumer_max = is_ann_note_value_defined_and_get_value(
                ann_note_consumer, self.max_field
            )
            producer_max_defined, producer_max = is_ann_note_value_defined_and_get_value(
                ann_note_producer, self.max_field
            )
            if consumer_max_defined:
                if not producer_max_defined:
                    self.violated = True
                    self.violation_text.append(f'{self.max_field}: {producer.name} is redefined.')
                elif float(consumer_max) < float(producer_max):
                    self.violated = True
                    self.violation_text.append(
                        f'{self.max_field}: {consumer.name} < {producer.name}.'
                    )


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    IdenticalForProducerConsumer()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoBoolComparison rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.apitools.query import get_leaf_type, is_predefined
from ansys.scade.design_rules.utils.naming import substitute_names
from ansys.scade.design_rules.utils.rule import Rule


class NoBoolComparison(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0054',
        category='Modelling',
        severity=Rule.MANDATORY,
        description='Boolean values should not be compared to the constants TRUE or FALSE.',
        label='No boolean comparison',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # resolution of internal_variables
        self.aliases = {}

        if self._search_through_expr_calls_const_bool(object_):
            # replace internal variables by their constants
            text = substitute_names(object_.to_string(), self.aliases)
            container = self.get_closest_annotatable(object_)
            error_msg = f'Bool comparison found ({text})'
            identifier = object_.to_string()
            self.add_rule_status(container, Rule.FAILED, error_msg, identifier)

        return Rule.NA

    def _search_through_expr_calls_const_bool(self, expr_call: suite.ExprCall) -> bool:
        # check only comparison operators
        if expr_call.predef_opr not in {20, 21, 22, 23, 24, 25}:
            return False

        operands = expr_call.parameters
        # go through all operands of the expression call
        for operand in operands:
            # resolve only internal variables, only once
            # ==> named constant flows are ignored
            if isinstance(operand, suite.ExprId):
                producer = operand.reference
                if isinstance(producer, suite.LocalVariable) and producer.is_internal():
                    # must be only one and only one definition
                    # assert len(producer.definitions) == 1
                    operand = producer.definitions[0].right
                    # replace the internal variable with its producer and store the alias
                    self.aliases[producer.name] = f'({operand.to_string()})'

            if isinstance(operand, suite.ConstValue):
                if operand.kind == 'Bool':
                    # one operand is a boolean constant, no need to continue
                    return True
            elif isinstance(operand, suite.ExprId):
                if isinstance(operand.reference, suite.Constant):
                    if self._is_bool(operand.reference.type):
                        # one operand is a boolean constant, no need to continue
                        return True

        return False

    def _is_bool(self, type_: suite.Type) -> bool:
        leaf_type = get_leaf_type(type_)
        return is_predefined(leaf_type) and leaf_type.name == 'bool'


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoBoolComparison()
